import scipy.sparse as spr
from sklearn.feature_extraction.text import CountVectorizer

from tests.sample import sample_texts
from tweetopic.dmm import DMM
from tweetopic.exceptions import NotFittedException

sample = CountVectorizer().fit_transform(sample_texts)
assert isinstance(sample, spr.csr_matrix)


def test_unfitted():
    dmm = DMM()
    assert dmm.components_ is None
    assert dmm.n_documents == 0
    assert dmm.cluster_doc_count is None
    assert dmm.n_features_in_ == 0
    assert not dmm._fitted
    is_error_raised = False
    try:
        dmm._check_fitted()
    except NotFittedException:
        is_error_raised = True
    assert is_error_raised, "NotFittedException is not raised"


def test_fitting():
    dmm = DMM().fit(sample)
    sample_dense = sample.toarray()
    DMM().fit(sample_dense)
    sample_lists = sample_dense.tolist()
    DMM().fit(sample_lists)
    assert dmm._fitted
    assert dmm.n_documents == len(sample_texts)
    assert dmm.n_features_in_ != 0
    assert dmm.components_ is not None
    assert dmm.components_.shape == (dmm.n_documents, dmm.n_features_in_)
    assert dmm.cluster_doc_count is not None


def test_params():
    params = {
        "n_components": 9,
        "n_iterations": 44,
        "alpha": 0.6,
        "beta": 0.5,
    }
    dmm = DMM()
    dmm.set_params(**params)
    assert dmm.get_params() == params


def test_shapes():
    dmm = DMM().fit(sample)
    transformed = dmm.transform(sample)
    n_transformed, n_features = transformed.shape
    assert n_transformed == len(sample_texts)
    assert n_features == dmm.n_features_in_
    prediction = dmm.predict(sample)
    (n_predicted,) = prediction.shape
    assert n_predicted == len(sample_texts)



"""Utilities for easier fitting, inspection and visualization of topic
models."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, Protocol, Union

import numpy as np
import scipy.sparse as spr
from numpy.typing import ArrayLike
from sklearn.decomposition import NMF, LatentDirichletAllocation

from tweetopic.dmm import DMM


class Vectorizer(Protocol):
    """Vectorizer protocol for static typing purposes."""

    def fit(self, raw_documents) -> Vectorizer:  # type: ignore
        pass

    def transform(self, raw_documents) -> Union[ArrayLike, spr.spmatrix]:  # type: ignore
        pass

    def fit_transform(self, raw_documents) -> Union[ArrayLike, spr.spmatrix]:  # type: ignore
        pass

    def get_feature_names_out(self) -> np.ndarray:  # type: ignore
        pass


TopicModel = Union[NMF, LatentDirichletAllocation, DMM]


@dataclass
class TopicPipeline:
    """Provides a full pipeline for topic modelling with useful utilities.

    Parameters
    ----------
    vectorizer: Vectorizer
        Transformer that extracts BOW embeddings from texts
    topic_model: DMM, NMF or LatentDirichletAllocation
        Topic model to add to the pipeline.
    """

    vectorizer: Vectorizer
    topic_model: TopicModel

    def fit(self, texts: Iterable[str]) -> TopicPipeline:
        """Fits vectorizer and topic model with the given stream of texts.

        Parameters
        ----------
        texts: iterable of str
            Stream of texts to fit the pipeline with.

        Returns
        -------
        TopicPipeline
            Fitted pipeline.
        """
        doc_term_matrix = self.vectorizer.fit_transform(texts)
        self.topic_model.fit(doc_term_matrix)
        return self

    def fit_transform(self, texts: Iterable[str]) -> np.ndarray:
        """Fits vectorizer and topic model and transforms the given text.

        Parameters
        ----------
        texts: iterable of str
            Texts to transform.

        Returns
        -------
        array of shape (n_documents, n_components)
            Document-topic matrix.
        """
        doc_term_matrix = self.vectorizer.fit_transform(texts)
        return self.topic_model.fit_transform(doc_term_matrix)

    def transform(self, texts: Iterable[str]) -> np.ndarray:
        """Transforms given texts with the fitted pipeline.

        Parameters
        ----------
        texts: iterable of str
            Texts to transform.

        Returns
        -------
        array of shape (n_documents, n_components)
            Document-topic matrix.
        """
        doc_term_matrix = self.vectorizer.transform(texts)
        return self.topic_model.transform(doc_term_matrix)

    def visualize(self, texts: Iterable[str]):
        """Visualizes the model with pyLDAvis for inspection of the different
        mixture components :)

        Parameters
        ----------
        texts: iterable of str
            Texts to visualize the model with.

        Raises
        ------
        ModuleNotFoundError
            If pyLDAvis is not installed an exception is thrown.
        """
        try:
            import pyLDAvis
            import pyLDAvis.sklearn
        except ModuleNotFoundError as exception:
            raise ImportError(
                "Optional dependency pyLDAvis not installed.",
            ) from exception
        pyLDAvis.enable_notebook()
        return pyLDAvis.sklearn.prepare(
            self.topic_model,
            self.vectorizer.transform(texts),
            self.vectorizer,
        )

    def visualise(self, texts: Iterable[str]):
        """Alias of :meth:`~tweetopic.pipeline.TopicPipeline.visualize` for
        those living on this side of the Pacific."""
        return self.visualize

    def top_words(self, top_n: int | None = 10) -> list[dict[str, int]]:
        """Calculates the top words for each cluster.

        Parameters
        ----------
        top_n: int or None, default 10
            Top N words to return. If None, all words are returned.

        Returns
        -------
        list of dict of str to int
            Dictionary for each cluster mapping the words to number of occurances.

        Raises
        ------
        NotFittedException
            If the model is not fitted, an exception will be raised.
        """
        feature_names = self.vectorizer.get_feature_names_out()
        dist: np.ndarray = self.topic_model.components_  # type: ignore
        res = []
        for i_cluster in range(self.topic_model.n_components):  # type: ignore
            top_indices = np.argsort(-dist[i_cluster])
            if top_n is not None:
                top_indices = top_indices[:top_n]  # type: ignore
            top_words = {
                feature_names[i]: dist[i_cluster, i]
                for i in top_indices
                if dist[
                    i_cluster,
                    i,
                ]  # Only return words if they are actually in the cluster
            }
            res.append(top_words)
        return res



sample_texts = [
    "dumbest person knows world controlled certain individuals richest",
    "e cigarettes cause horrible incurable disease called popcorn lung",
    "possible reach state peace harmony hands",
    "brain releases neurotransmitter dopamine crucial numerous essential bodily functions",
    "following list characteristics regularly manifested reptilian possessed reptilian hybrid",
    "ancient site dates years hides secret messages visible night",
    "time joey verwey years old telling family amazing stories able",
    "humans aliens looking according experts humans likely",
    "prediction found book revelations emerald tablets forecasting",
    "nazca civilization thrived bc ad arid plains southern peru nazca",
]



"""Module containing tools for fitting a Dirichlet Mixture Model."""
from __future__ import annotations

import numpy as np
from numba import njit

from tweetopic._doc import add_doc, remove_doc
from tweetopic._prob import predict_doc, sample_categorical


@njit
def init_clusters(
    cluster_word_distribution: np.ndarray,
    cluster_word_count: np.ndarray,
    cluster_doc_count: np.ndarray,
    doc_clusters: np.ndarray,
    doc_unique_words: np.ndarray,
    doc_unique_word_counts: np.ndarray,
    max_unique_words: int,
) -> None:
    """Randomly initializes clusters in the model.

    Parameters
    ----------
    cluster_word_count(OUT): array of shape (n_clusters,)
        Contains the amount of words there are in each cluster.
    cluster_word_distribution(OUT): matrix of shape (n_clusters, n_vocab)
        Contains the amount a word occurs in a certain cluster.
    cluster_doc_count(OUT): array of shape (n_clusters,)
        Array containing how many documents there are in each cluster.
    doc_clusters: array of shape (n_docs)
        Contains a cluster label for each document, that has
        to be assigned.
    doc_unique_words: matrix of shape (n_documents, MAX_UNIQUE_WORDS)
        Matrix containing all indices of unique words in the document.
    doc_unique_word_counts: matrix of shape (n_documents, MAX_UNIQUE_WORDS)
        Matrix containing all counts for each unique word in the document.
    max_unique_words: int
        Maximum number of unique words in any document.

    NOTE
    ----
    Beware that the function modifies a numpy array, that's passed in as
    an input parameter. Should not be used in parallel, as race conditions
    might arise.
    """
    n_docs, _ = doc_unique_words.shape
    for i_doc in range(n_docs):
        i_cluster = doc_clusters[i_doc]
        add_doc(
            i_doc=i_doc,
            i_cluster=i_cluster,
            cluster_word_distribution=cluster_word_distribution,
            cluster_word_count=cluster_word_count,
            cluster_doc_count=cluster_doc_count,
            doc_unique_words=doc_unique_words,
            doc_unique_word_counts=doc_unique_word_counts,
            max_unique_words=max_unique_words,
        )


@njit(parallel=False)
def fit_model(
    n_iter: int,
    alpha: float,
    beta: float,
    n_clusters: int,
    n_vocab: int,
    n_docs: int,
    doc_unique_words: np.ndarray,
    doc_unique_word_counts: np.ndarray,
    doc_clusters: np.ndarray,
    cluster_doc_count: np.ndarray,
    cluster_word_count: np.ndarray,
    cluster_word_distribution: np.ndarray,
    max_unique_words: int,
) -> None:
    """Fits the Dirichlet Mixture Model with Gibbs Sampling. Implements
    algorithm described in Yin & Wang (2014)

    Parameters
    ----------
    n_iter: int
        Number of iterations to conduct.
    alpha: float
        Alpha parameter of the model.
    beta: float
        Beta parameter of the model.
    n_clusters: int
        Number of mixture components in the model.
    n_vocab: int
        Number of total vocabulary items.
    n_docs: int
        Total number of documents.
    doc_term_matrix: matrix of shape (n_documents, n_vocab)
        Contains how many times a term occurs in each document.
        (Bag of words matrix)
    doc_unique_words: array of shape (MAX_UNIQUE_WORDS, )
        Array containing all the ids of unique words in a document.
    doc_clusters: array of shape (n_docs, )
        Contains a cluster label for each document, that has
        to be assigned.
    doc_unique_words_count: array of shape (n_documents, )
        Vector containing the number of unique terms in each document.
    cluster_doc_count(OUT): array of shape (n_clusters,)
        Array containing how many documents there are in each cluster.
    cluster_word_count(OUT): array of shape (n_clusters,)
        Contains the amount of words there are in each cluster.
    cluster_word_distribution(OUT): matrix of shape (n_clusters, n_vocab)
        Contains the amount a word occurs in a certain cluster.
    max_unique_words: int
        Maximum count of unique words in a document seen in the corpus.
    """
    doc_word_count = np.sum(doc_unique_word_counts, axis=1)
    prediction = np.empty(n_clusters)
    for iteration in range(n_iter):
        total_transfers = 0
        for i_doc in range(n_docs):
            # Removing document from previous cluster
            prev_cluster = doc_clusters[i_doc]
            # Removing document from the previous cluster
            remove_doc(
                i_doc,
                prev_cluster,
                cluster_word_distribution=cluster_word_distribution,
                cluster_word_count=cluster_word_count,
                cluster_doc_count=cluster_doc_count,
                doc_unique_words=doc_unique_words,
                doc_unique_word_counts=doc_unique_word_counts,
                max_unique_words=max_unique_words,
            )
            # Getting new prediction for the document at hand
            predict_doc(
                probabilities=prediction,
                i_document=i_doc,
                doc_unique_words=doc_unique_words,
                doc_unique_word_counts=doc_unique_word_counts,
                n_words=doc_word_count[i_doc],
                alpha=alpha,
                beta=beta,
                n_clusters=n_clusters,
                n_vocab=n_vocab,
                n_docs=n_docs,
                cluster_doc_count=cluster_doc_count,
                cluster_word_count=cluster_word_count,
                cluster_word_distribution=cluster_word_distribution,
                max_unique_words=max_unique_words,
            )
            new_cluster = sample_categorical(prediction)
            if prev_cluster != new_cluster:
                total_transfers += 1
            # Adding document back to the newly chosen cluster
            doc_clusters[i_doc] = new_cluster
            add_doc(
                i_doc,
                new_cluster,
                cluster_word_distribution=cluster_word_distribution,
                cluster_word_count=cluster_word_count,
                cluster_doc_count=cluster_doc_count,
                doc_unique_words=doc_unique_words,
                doc_unique_word_counts=doc_unique_word_counts,
                max_unique_words=max_unique_words,
            )
        n_populated = np.count_nonzero(cluster_doc_count)
        print(
            f" Iteration {iteration}/{n_iter}: transferred {total_transfers} documents, {n_populated} clusters remain populated.",
        )
